
#ifndef ___xfdashboard_marshal_MARSHAL_H__
#define ___xfdashboard_marshal_MARSHAL_H__

#include	<glib-object.h>

G_BEGIN_DECLS

/* VOID:FLOAT,FLOAT (marshal.list:1) */
extern void _xfdashboard_marshal_VOID__FLOAT_FLOAT (GClosure     *closure,
                                                    GValue       *return_value,
                                                    guint         n_param_values,
                                                    const GValue *param_values,
                                                    gpointer      invocation_hint,
                                                    gpointer      marshal_data);

/* VOID:INT,INT (marshal.list:2) */
extern void _xfdashboard_marshal_VOID__INT_INT (GClosure     *closure,
                                                GValue       *return_value,
                                                guint         n_param_values,
                                                const GValue *param_values,
                                                gpointer      invocation_hint,
                                                gpointer      marshal_data);

/* VOID:INT,INT,INT,INT (marshal.list:3) */
extern void _xfdashboard_marshal_VOID__INT_INT_INT_INT (GClosure     *closure,
                                                        GValue       *return_value,
                                                        guint         n_param_values,
                                                        const GValue *param_values,
                                                        gpointer      invocation_hint,
                                                        gpointer      marshal_data);

/* VOID:OBJECT,OBJECT (marshal.list:4) */
extern void _xfdashboard_marshal_VOID__OBJECT_OBJECT (GClosure     *closure,
                                                      GValue       *return_value,
                                                      guint         n_param_values,
                                                      const GValue *param_values,
                                                      gpointer      invocation_hint,
                                                      gpointer      marshal_data);

/* VOID:OBJECT,FLOAT,FLOAT (marshal.list:5) */
extern void _xfdashboard_marshal_VOID__OBJECT_FLOAT_FLOAT (GClosure     *closure,
                                                           GValue       *return_value,
                                                           guint         n_param_values,
                                                           const GValue *param_values,
                                                           gpointer      invocation_hint,
                                                           gpointer      marshal_data);

/* VOID:OBJECT,OBJECT,OBJECT (marshal.list:6) */
extern void _xfdashboard_marshal_VOID__OBJECT_OBJECT_OBJECT (GClosure     *closure,
                                                             GValue       *return_value,
                                                             guint         n_param_values,
                                                             const GValue *param_values,
                                                             gpointer      invocation_hint,
                                                             gpointer      marshal_data);

/* VOID:STRING,BOOLEAN (marshal.list:7) */
extern void _xfdashboard_marshal_VOID__STRING_BOOLEAN (GClosure     *closure,
                                                       GValue       *return_value,
                                                       guint         n_param_values,
                                                       const GValue *param_values,
                                                       gpointer      invocation_hint,
                                                       gpointer      marshal_data);

/* BOOLEAN:OBJECT (marshal.list:8) */
extern void _xfdashboard_marshal_BOOLEAN__OBJECT (GClosure     *closure,
                                                  GValue       *return_value,
                                                  guint         n_param_values,
                                                  const GValue *param_values,
                                                  gpointer      invocation_hint,
                                                  gpointer      marshal_data);

/* BOOLEAN:OBJECT,ENUM (marshal.list:9) */
extern void _xfdashboard_marshal_BOOLEAN__OBJECT_ENUM (GClosure     *closure,
                                                       GValue       *return_value,
                                                       guint         n_param_values,
                                                       const GValue *param_values,
                                                       gpointer      invocation_hint,
                                                       gpointer      marshal_data);

/* BOOLEAN:OBJECT,FLOAT,FLOAT (marshal.list:10) */
extern void _xfdashboard_marshal_BOOLEAN__OBJECT_FLOAT_FLOAT (GClosure     *closure,
                                                              GValue       *return_value,
                                                              guint         n_param_values,
                                                              const GValue *param_values,
                                                              gpointer      invocation_hint,
                                                              gpointer      marshal_data);

/* BOOLEAN:OBJECT,STRING,OBJECT (marshal.list:11) */
extern void _xfdashboard_marshal_BOOLEAN__OBJECT_STRING_OBJECT (GClosure     *closure,
                                                                GValue       *return_value,
                                                                guint         n_param_values,
                                                                const GValue *param_values,
                                                                gpointer      invocation_hint,
                                                                gpointer      marshal_data);

/* BOOLEAN:VOID (marshal.list:12) */
extern void _xfdashboard_marshal_BOOLEAN__VOID (GClosure     *closure,
                                                GValue       *return_value,
                                                guint         n_param_values,
                                                const GValue *param_values,
                                                gpointer      invocation_hint,
                                                gpointer      marshal_data);

/* OBJECT:VOID (marshal.list:13) */
extern void _xfdashboard_marshal_OBJECT__VOID (GClosure     *closure,
                                               GValue       *return_value,
                                               guint         n_param_values,
                                               const GValue *param_values,
                                               gpointer      invocation_hint,
                                               gpointer      marshal_data);

G_END_DECLS

#endif /* ___xfdashboard_marshal_MARSHAL_H__ */

